/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Charts module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:GPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 or (at your option) any later version
** approved by the KDE Free Qt Foundation. The licenses are as published by
** the Free Software Foundation and appearing in the file LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "datasourcescope.h"
#include <iostream>
#include <QtCharts/QXYSeries>
#include <QtCharts/QAreaSeries>
#include <QtQuick/QQuickView>
#include <QtQuick/QQuickItem>
#include <QtCore/QDebug>
#include <QtCore/QtMath>

QT_CHARTS_USE_NAMESPACE

Q_DECLARE_METATYPE(QAbstractSeries *)
Q_DECLARE_METATYPE(QAbstractAxis *)

DataSourceScope::DataSourceScope(QQuickView *appViewer, QObject *parent) :
    QObject(parent),
    m_appViewer(appViewer),
    m_index(-1)
{
    qRegisterMetaType<QAbstractSeries*>();
    qRegisterMetaType<QAbstractAxis*>();

     // Create an interface to the 6022be scope
     _scopeInterface = new ScopeInterface();

     // Create a buffer for each channel (max 2 at the moment).
     _channels.append( new Channel());
     _channels.append( new Channel());

     if( ! _scopeInterface->setup()) {
          cerr << "Cannot setup hantek 6022be." << endl;
          // exit(1);
     }

     if( _scopeInterface->getScopeStatus()==is_setup && !_scopeInterface->start_capture()) {
        cerr << "Cannot start sampling." << endl;
        // exit(1);
     }

    cout << "DataSource constructor complete\n";

    if( _scopeInterface->getScopeStatus()==capturing) {
        readData(2, 8192);
    }
}

DataSourceScope::~DataSourceScope() {
    if( _scopeInterface) {
        delete _scopeInterface;
    }
}

/**
 * @brief DataSource::getNumChannels gets the number of channels.
 *
 * @return The current number of channels.
 */
int DataSourceScope::getNumChannels() {
    return 2;  // Hantek firmware has always 2 channels.
}

void DataSourceScope::update(QAbstractSeries *series, int channel_index)
{
    if (series) {

        // Update all the data, while channel 0 is fetched
        if( channel_index == 0) {
            readData( 2, 1024);
        }

        QXYSeries *xySeries = static_cast<QXYSeries *>(series);

        QVector<QPointF> points = _channels.at(channel_index)->getCurrentBuffer(NULL,100);

        // Use replace instead of clear + append, it's optimized for performance
        xySeries->replace(points);

    }
}

void DataSourceScope::readData(int channelCount, int colCount)
{

    cout << "readData started\n";

    // Remove previous data
    foreach (QVector<QPointF> row, m_data) {
        row.clear();
    }
    m_data.clear();

    if( _scopeInterface->getScopeStatus() == capturing) {

        // Fetch data from the scope
        ScopeData scopeData;
        if( _scopeInterface->read_data( 2, 8192, &scopeData)) {

            cout << "Data from scope fetched\n";

            // Append the new data depending on the type
            for (int i(0); i < scopeData._nChannels; i++) {
                //QVector<QPointF> points;
                //points.reserve(colCount);
                for (int j(0); j < scopeData._length; j++) {
                    qreal x(0);
                    qreal y(0);
                    y=scopeData._buffers[i][j] * 5.0 / 255.0;
                    x=j;

                    //points.append(QPointF(x, y));  // <= remove later!
                    _channels.at(i)->addPoint(QPointF(x,y));
                }
                //m_data.append(points);
            }
        } else {  // Just append a 0 line?
            cout << "Scope not capturing\n";
        }
    } else {
        cerr << "Scope is not capturing. Cannot read data from it." << endl;
    }
}
