#include "trigger.h"


/**
 * @brief Trigger::Trigger creates a new software trigger.
 */
Trigger::Trigger() {
}

/**
 * @brief Trigger::Trigger creates a new trigger type with a given type and voltage.
 *
 * @param type The type of the trigger.
 * @param volt The trigger voltage.
 */
Trigger::Trigger( TriggerType type, float volt) {
    _type = type;
    setVoltage( volt);
}


/**
 * @brief Trigger::findTriggerPos  tries to find the index, where a trigger fires in the channel data
 *
 * @param channelData The data of the scope channel as a vector of points.
 * @param start The start index of the search
 * @param end The end index of the search
 *
 * @return The index of the point, where the trigger fires or -1, if it never fires in the current channel data.
 */
int Trigger::findTriggerPos( QVector<QPointF> channelData, int start, int end) {

    if( _type == POS_EDGE) {
        int prev_index = (start == 0 ? channelData.size()-1 : start - 1);
        for( int i=start+1; i != end; i = ( i == channelData.size() - 1 ? 0 : i + 1)) {
            if( (channelData.at( prev_index).y() < _triggerVoltage)
                && (channelData.at(i).y() >= _triggerVoltage)) {
                  return i;
            }
            prev_index = i;
        }

    } else if( _type == NEG_EDGE) {

        int prev_index = (start == 0 ? channelData.size()-1 : start - 1);
        for( int i=start+1; i != end; i = ( i == channelData.size() - 1 ? 0 : i + 1)) {
            if( (channelData.at( prev_index).y() > _triggerVoltage)
                && (channelData.at(i).y() <= _triggerVoltage)) {
                  return i;
            }
            prev_index = i;
        }
    }

    return -1;  // Trigger does not fire in this data.
}

/**
 * @brief Trigger::setVoltage sets a new voltage for the trigger level
 *
 * @param volt The new trigger level.
 */
void Trigger::setVoltage(float volt) {
    _triggerVoltage = volt;
}
