/**
  * Interface for usb oscilloscope hantek 6022be
  *
  * @author Andreas Rueckert <arueckert67@t-online.de>
  *
  * This is a port of this code: https://github.com/rpcope1/Hantek6022API
  *
  */

#include "scopeinterface.h"

#include <assert.h>
#include <fstream>
#include <iostream>
#include <libusb-1.0/libusb.h>
#include <regex>
#include <string>
#include <unistd.h>

using namespace std;

#define VENDOR_ID_NO_FIRMWARE 0x04b4
#define VENDOR_ID 0x04b5
#define PRODUCT_ID 0x6022

// Defines for usb operations
#define RW_FIRMWARE_REQUEST 0xa0  // Send a firmware packet
#define RW_FIRMWARE_INDEX   0x00

#define SET_CH_VR_REQUEST   0xe0
#define SET_CH_VR_VALUE     0x00
#define SET_CH_VR_INDEX     0x00

#define SET_SAMPLE_RATE_REQUEST 0xe2
#define SET_SAMPLE_RATE_VALUE   0x00
#define SET_SAMPLE_RATE_INDEX   0x00

#define TRIGGER_REQUEST     0xe3
#define TRIGGER_VALUE       0x00
#define TRIGGER_INDEX       0x00

#define DEBUG 1


/**
 * @brief ScopeInterface::ScopeInterface
 *
 * @see <a href="http://www.dreamincode.net/forums/topic/148707-introduction-to-using-libusb">libusb introduction</a>
 */
ScopeInterface::ScopeInterface() {
    /* libusb_device  **devs;            // pointer to libusb devices
    ssize_t          device_count;    // The number of devices. */

    // Get the list of available devices.
    /* device_count = libusb_get_device_list( _context, &devs);

    cout << "Found " << device_count << " usb devices";

    ssize_t i;
    for( i = 0; i < device_count; ++i) {

        libusb_device_descriptor desc;

        int return_val = libusb_get_device_descriptor( devs[i], &desc);

        if( return_val < 0) {

            cerr << "Failed to get device descriptor" << endl;
            return;
        }

        // Check for a hantek 6022 usb scope without firmware.
        if( desc.idVendor == 0x04b4 && desc.idProduct == 0x6022) {

            cout << "Found hantek 6022 scope without loaded firmware." << endl;
        }

        if( desc.idVendor == 0x04b5 && desc.idProduct == 0x6022) {

            cout << "Found hantek 6022 scope with loaded firmware." << endl;
        }
    }

    // Free all devices for now.
    libusb_free_device_list( devs, 1);
    libusb_exit( context); */

    _context = NULL;

    // Create a buffer for faster reading.
    _read_buffer = new unsigned char[65536];
}

/**
 * @brief ScopeInterface::~ScopeInterface cleanly close the connection to the scope.
 */
ScopeInterface::~ScopeInterface() {

    if( _scope_device_handle != NULL) {
        libusb_close(_scope_device_handle);
    }

    libusb_exit(_context);  // Close the usb connection.
}

/**
 * Flash a firmware via USB to the scope.
 *
 * @return true, if the firmware was successfully flashed.
 */
bool ScopeInterface::flash_firmware() {

    vector<UsbFirmwarePacket> firmware = read_firmware( "/home/andreas/hantek_6022/qtscope/QtScope/firmware/hantek.fw");

    for( size_t i=0; i < firmware.size(); ++i) {

        UsbFirmwarePacket currentPacket = firmware[i];

        if( libusb_control_transfer( _scope_device_handle
                                     , 0x40
                                     , RW_FIRMWARE_REQUEST
                                     , currentPacket.value
                                     , RW_FIRMWARE_INDEX
                                     , (unsigned char *)currentPacket.data.data()
                                     , currentPacket.data.length()
                                     , 60000) != currentPacket.data.length()) {

                // Error: not all of the firmware bytes were transfered.
                return false;
        }
    }

    return true;  // Success!
}

/**
 * @brief ScopeInterface::getScopeStatus returns the current status of the scope.
 *
 * @return The current status of the scope.
 */
scopeStatus ScopeInterface::getScopeStatus() {
    return _scopeStatus;
}

/**
  * List available usb devices. Mainly useful for debugging.
  *
  * @see https://gist.github.com/sparkbuzz/7695955
  */
void ScopeInterface::list_usb_devices() {
    ssize_t device_count = 0;
    libusb_device **usb_devices = NULL;

    // Get USB device list
    device_count = libusb_get_device_list( _context, &usb_devices);
    if( device_count < 0) {
        cerr << "Unable to retrieve USB device list!\n";
     }

     // Iterate and print devices
     cout << ("VID    PID     Manufacturer Name\n------ ------ -------------------\n");
     for( int i = 0; i < device_count; ++i) {

         struct libusb_device_descriptor device_descriptor;
         struct libusb_device_handle *device_handle = NULL;

         // Get USB device descriptor
         int result = libusb_get_device_descriptor( usb_devices[i], &device_descriptor);
         if( result < 0) {
            cerr << "Failed to get device descriptor!\n";
         }

         // Print VID & PID
         printf("0x%04x 0x%04x\n", device_descriptor.idVendor, device_descriptor.idProduct);
     }

     libusb_free_device_list(usb_devices,1);
}


/**
 * @brief Open a handle to the scope.
 *
 * @return true, if the handle was successfully opened and store in a instance variable.
 */
bool ScopeInterface::open_handle() {

    // If there is no handle yet, setup the connection.
    if( _scope_device_handle == NULL) {
        setup();
    }

    if( _scope_device_handle == NULL) {
        return false;
    }

    if( 0 == libusb_claim_interface( _scope_device_handle, 0)) {
        return true;
    }

    return false;  // Opening the handle failed.
}


/**
 * @brief read_data fetches the data from the active scope channels.
 *
 * @param nChannels The number of channals to read.
 * @param length The buffer length of each channel.
 * @param result A buffer for the returned data.
 *
 * @return true, if the data were successfully read. False otherwise.
 */
bool ScopeInterface::read_data( int nChannels, int length, ScopeData *result) {

    int transfered_bytes;
    result->_nChannels = nChannels;
    result->_length = length;
    result->_buffers = new unsigned char *[nChannels];

    // The following code is slooow! Don't create a new array for every scope reading!
    for( int i = 0; i < nChannels; ++i) {
        result->_buffers[ i] = new unsigned char[ length];
    }

    int retVal;
    if( ( retVal = libusb_bulk_transfer( _scope_device_handle
                          , 0x86
                          , _read_buffer
                          , nChannels * length
                          , &transfered_bytes
                          , 60000)) != 0UL ) {

        cerr << "ERROR: cannot read captured data. Error: " << retVal << endl;

        return false;
    }

    // Set the transferred byte into the scope data.
    for( int i=0; i < nChannels; ++i) {
        result->_length = transfered_bytes / nChannels;

        // Copy the bytes to the channel data buffer
        unsigned int source_index = i;
        unsigned dest_index = 0;
        while( source_index < transfered_bytes) {
            result->_buffers[i][dest_index++] = _read_buffer[source_index];
            source_index += nChannels;
        }
    }

    // Fetching worked.
    return true;
}

vector<UsbFirmwarePacket> ScopeInterface::read_firmware( string firmware) {

    // Check the filename, if this is an ihex file.
    string suffix = string( ".ihx");

    size_t suffix_len = suffix.length();
    size_t filename_len = firmware.length();

    // If the filename ends with an ihex suffix.
    if( ( suffix_len <= filename_len)
            && ( string::npos != firmware.rfind( suffix.c_str()
                                                 , filename_len - suffix_len
                                                 ,  suffix_len))) {

        // Try to read ihex file.
        return read_firmware_ihex( firmware);
    }

    // Try to read CSV file.
    return read_firmware_csv( firmware);
}

/**
 * @brief ScopeInterface::read_firmware reads a firmware from a CSV file.
 *
 * @param filename The name of the file holding the firmware data.
 *
 * @return The firmware as USB packages.
 */
vector<UsbFirmwarePacket> ScopeInterface::read_firmware_csv( string filename) {

    vector<UsbFirmwarePacket> firmwareBuffer;

    // If there is no firmware filename given, use the default.
    if( filename.empty()) {
        filename = "oscilloscope/ht6022be/firmware/hantek.fw";
    }

    ifstream firmwareFile;

    firmwareFile.open( filename.c_str(), ifstream::in);
    if( ! firmwareFile.is_open()) {

        cerr << "Cannot open firmware file "
             << filename
             << endl;

        firmwareFile.close();
        return firmwareBuffer;
    }

    // Turn 8051 off, before flashing the firmware
    UsbFirmwarePacket newPacket;
    newPacket.value = 0xe600;
    newPacket.data = '\x01';
    firmwareBuffer.push_back( newPacket);

    // Now read the actual file.
    string currentLine;
    while( getline( firmwareFile, currentLine)) {

        // Trim the read line.
        currentLine.assign( regex_replace( currentLine.c_str(), regex("^ +| +$|( ) +"), "$1"));

        // If this is not a empty line.
        if( currentLine.size() > 0) {

            // Split the line.
            unsigned int dataLen = strtoul(strtok(&currentLine[0], ","),NULL,16);
            unsigned int value = strtoul(strtok(NULL,","),NULL,16);
            string data;
            for( unsigned int currentData = 0; currentData < dataLen; ++currentData) {

                data.append( sizeof(char), (char)(strtoul(strtok(NULL,","),NULL,16) & 0xff));
            }

            // Now add the current data to the firmware buffer.
            UsbFirmwarePacket newPacket2;
            newPacket2.value = value;
            newPacket2.data = data;
            firmwareBuffer.push_back( newPacket2);
        }
    }

    // Turn 8051 on again.
    UsbFirmwarePacket newPacket3;
    newPacket3.value = 0xe600;
    newPacket3.data = '\x00';
    firmwareBuffer.push_back( newPacket3);

    return firmwareBuffer;
}

bool ScopeInterface::setup() {

#ifdef DEBUG
    cout << "DEBUG: start setup" << endl;
#endif

    int return_value = libusb_init( &_context);  // Init the lib.

    if( return_value != 0) {
        cerr << "Cannot init libusb: " << return_value << endl;
        return false;
    }

#ifdef DEBUG
    cout << "DEBUG: libusb initialized" << endl;
#endif

    libusb_set_debug( _context, 3);  // Set debug level to 3.

    // Look for a hantek 6022 without loaded firmware.
    _scope_device_handle = libusb_open_device_with_vid_pid( _context
                                                            , VENDOR_ID_NO_FIRMWARE
                                                            , PRODUCT_ID);

    // If there is a device, try to flash the firmware.
    if( _scope_device_handle != NULL) {

        _scopeStatus = found;

        // If flashing the firmware doesn't work,
        if( ! flash_firmware()) {

#ifdef DEBUG
            cout << "ERROR: cannot flash firmware" << endl;
#endif

            return false;  // return an error.
        } else {
            _scopeStatus = flashed;
        }
    } else {
#ifdef DEBUG
        cout << "No hantek 6022be without firmware found" << endl;
#endif
    }

    // Maybe the scope already has it's firmware flashed?

    // Re-init context.
    if( int retVal = libusb_init( &_context) != 0) {

        cerr << "DEBUG: cannot re-init context, Returned " << retVal << endl;

        return false;
    }

    // Wait a bit for the scope to reinit.
    usleep( 2000000L);

    // Now try to get a hantek 6022 with a flashed firmware.
    _scope_device_handle = libusb_open_device_with_vid_pid( _context
                                                            , VENDOR_ID
                                                            , PRODUCT_ID);

    if( _scope_device_handle == NULL) {
        cerr << "Cannot get usb handle for flashed 6022be\n";
        list_usb_devices();
        return false;
    }

    _scopeStatus = flashed;

    if( _scope_device_handle != NULL) {
        _scopeStatus = is_setup;
        return true;
    }

    return false;
}

/**
 * @brief ScopeInterface::read_firmware_ihex reads a firmware from a ihex file.
 *
 * @param filename The name of the file holding the ihex data.
 *
 * @return The firmware as USB packages.
 */
vector<UsbFirmwarePacket> ScopeInterface::read_firmware_ihex( string filename) {

    vector<UsbFirmwarePacket> firmwareBuffer;

    // If there is no firmware filename given, use the default.
    if( filename.empty()) {
        filename = "firmware/jh.ihx";
    }

    ifstream firmwareFile;

    firmwareFile.open( filename.c_str(), ifstream::in);
    if( ! firmwareFile.is_open()) {

        cerr << "Cannot open firmware file "
             << filename
             << endl;

        firmwareFile.close();
        return firmwareBuffer;
    }

    // Turn 8051 off, before flashing the firmware
    UsbFirmwarePacket newPacket;
    newPacket.value = 0xe600;
    newPacket.data = '\x01';
    firmwareBuffer.push_back( newPacket);

    // Now read the actual file.
    string currentLine;
    while( getline( firmwareFile, currentLine)) {

        // Trim the read line.
        currentLine.assign( regex_replace( currentLine.c_str(), regex("^ +| +$|( ) +"), "$1"));

        // If this is not a empty line.
        if( currentLine.size() > 0) {

            assert( currentLine[0] == ':');

            unsigned int record_len = std::stoul( currentLine.substr( 1, 2), nullptr, 16);
            unsigned int addr = std::stoul( currentLine.substr( 3, 4), nullptr, 16);
            unsigned int record_type = std::stoul( currentLine.substr( 7,2), nullptr, 16);
            string data;
            for( unsigned int currentData = 0; currentData < record_len; ++currentData) {

                data.append( sizeof(char), (char)(std::strtoul( currentLine.substr( 9 + currentData * 2, 2).c_str(),nullptr,16) & 0xff));
            }
            unsigned int file_checksum = std::stoul( currentLine.substr( currentLine.size() - 2, 2), nullptr, 16);

            if( record_type == 0x00) {

                vector<unsigned char> dataArray;

                // Calculate a checksum from the current line.
                int checksum = 0;
                for( int i=9; i < currentLine.size()-2; i += 2) {

                    // Take 2 chars as a hex number. convert it to int
                    // and add it to the checksum.
                    int cur_number = std::stoul( currentLine.substr( i,2).c_str(), nullptr, 16);
                    checksum += cur_number;

                    dataArray.push_back( (unsigned char)cur_number);
                }
                checksum += record_len;
                checksum += addr & 0xff;
                checksum += addr >> 8;
                checksum &= 0xff;

                assert( (~((checksum + file_checksum) & 0xff) & 0xff) == 0);

                UsbFirmwarePacket newPacket;
                newPacket.value = addr;
                char dataString[ dataArray.size()];
                for( int i = 0; i < dataArray.size(); ++i) {
                    dataString[i] = dataArray[i];
                }
                newPacket.data = dataString;
                firmwareBuffer.push_back( newPacket);

            } else if( record_type == 0x01) {

                assert( file_checksum == 0xff);
                break;

            } else {

                cerr << "Unknown record type found: " << record_type << endl;
                firmwareBuffer.clear();  // Remove cpu stop.
                return firmwareBuffer;  // And return empty buffer.
            }

            // Turn 8051 on again.
            UsbFirmwarePacket newPacket3;
            newPacket3.value = 0xe600;
            newPacket3.data = '\x00';
            firmwareBuffer.push_back( newPacket3);

            return firmwareBuffer;
        }
    }
}

/**
 * @brief ScopeInterface::set_sample_rate sets the new sample rate for a given channel.
 * @param rate_code The code for the new sample rate.
 *
 * @return True, if the sample rate was successfully set. False otherwise.
 */
bool ScopeInterface::set_sample_rate( unsigned char rate_code) {

    return libusb_control_transfer( _scope_device_handle
                                      , 0x40
                                      , SET_SAMPLE_RATE_REQUEST
                                      , SET_SAMPLE_RATE_VALUE
                                      , SET_SAMPLE_RATE_INDEX
                                      , &rate_code
                                      , 1
                                      , 60000) == 1;
}

/**
 * @brief ScopeInterface::set_voltage_range Set the voltage range of a channel.
 * @param channel The channel to set (0 or 1).
 * @param range_code A byte indicating the selected range.
 *
 * @return true, if the range was successsfully selected.
 */
bool ScopeInterface::set_voltage_range( int channel,
                                        unsigned char range_code) {
    assert( channel < _nActiveChannels);
    assert( _scope_device_handle != NULL);

    return libusb_control_transfer( _scope_device_handle
                                      , 0x40
                                      , SET_CH_VR_REQUEST + channel
                                      , SET_CH_VR_VALUE
                                      , SET_CH_VR_INDEX
                                      , &range_code
                                      , 1
                                      , 60000) == 1;

}

/**
 * @brief tell the scope to start capturing data.
 *
 * @return true, if the scope accepts the command. False otherwise.
 */
bool ScopeInterface::start_capture() {

    cout << "DEBUG: start capturing" << endl;

    if( _scope_device_handle == NULL) {

        cout << "DEBUG: cannot start capturing. No device handle" << endl;

        return false;
    }

    unsigned char data = '0x00';
    if( libusb_control_transfer( _scope_device_handle
                                      , 0x40
                                      , TRIGGER_REQUEST
                                      , TRIGGER_VALUE
                                      , TRIGGER_INDEX
                                      , &data
                                      , 1
                                      , 60000) == 1) {
        _scopeStatus = capturing;
        return true;
    }
    return false;
}

/**
 * @brief tell the scope to stop capturing data. This is only supported with the
 *        custom firmware.
 *
 * @return true, if the scope accepts the stop command. False otherwise.
 */
bool ScopeInterface::stop_capture() {

    unsigned char data = 0x00;
    if ( libusb_control_transfer( _scope_device_handle
                                      , 0x40
                                      , TRIGGER_REQUEST
                                      , TRIGGER_VALUE
                                      , TRIGGER_INDEX
                                      , &data
                                      , 1
                                      , 60000) == 1) {
        _scopeStatus = is_setup;
        return true;
    }
    return false;
}

